// =============================================================================
//
// ztd.text
// Copyright © 2021 JeanHeyd "ThePhD" Meneide and Shepherd's Oasis, LLC
// Contact: opensource@soasis.org
//
// Commercial License Usage
// Licensees holding valid commercial ztd.text licenses may use this file in
// accordance with the commercial license agreement provided with the
// Software or, alternatively, in accordance with the terms contained in
// a written agreement between you and Shepherd's Oasis, LLC.
// For licensing terms and conditions see your agreement. For
// further information contact opensource@soasis.org.
//
// Apache License Version 2 Usage
// Alternatively, this file may be used under the terms of Apache License
// Version 2.0 (the "License") for non-commercial use; you may not use this
// file except in compliance with the License. You may obtain a copy of the 
// License at
//
//		http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//
// ============================================================================>

#pragma once

#ifndef ZTD_TEXT_IS_LOSSLESS_HPP
#define ZTD_TEXT_IS_LOSSLESS_HPP

#include <ztd/text/version.hpp>

#include <ztd/text/error_handler.hpp>
#include <ztd/text/is_full_range_representable.hpp>

#include <ztd/text/detail/type_traits.hpp>

#include <type_traits>

namespace ztd { namespace text {
	ZTD_TEXT_INLINE_ABI_NAMESPACE_OPEN_I_

	namespace __txt_detail {
		template <typename _Encoding, typename _ErrorHandler>
		class __is_encode_lossless_or_deliberate
		: public ::std::integral_constant<bool,
			  __txt_detail::__is_careless_error_handler_v<_ErrorHandler> ? is_encode_injective_v<_Encoding> : true> { };

		template <typename _Encoding, typename _ErrorHandler>
		inline constexpr bool __is_encode_lossless_or_deliberate_v
			= __is_encode_lossless_or_deliberate<_Encoding, _ErrorHandler>::value;

		template <typename _Encoding, typename _ErrorHandler>
		class __is_decode_lossless_or_deliberate
		: public ::std::integral_constant<bool,
			  __txt_detail::__is_careless_error_handler_v<_ErrorHandler> ? is_decode_injective_v<_Encoding> : true> { };

		template <typename _Encoding, typename _ErrorHandler>
		inline constexpr bool __is_decode_lossless_or_deliberate_v
			= __is_decode_lossless_or_deliberate<_Encoding, _ErrorHandler>::value;

		template <typename _FromEncoding, typename _ToEncoding, typename _ErrorHandler>
		class __is_transcode_lossless_or_deliberate
		: public ::std::integral_constant<bool,
			  __txt_detail::__is_careless_error_handler_v<_ErrorHandler>
			       ? is_decode_injective_v<_FromEncoding> && is_encode_injective_v<_ToEncoding>
			       : true> { };

		template <typename _FromEncoding, typename _ToEncoding, typename _ErrorHandler>
		inline constexpr bool __is_transcode_lossless_or_deliberate_v
			= __is_transcode_lossless_or_deliberate<_FromEncoding, _ToEncoding, _ErrorHandler>::value;
	} // namespace __txt_detail

	ZTD_TEXT_INLINE_ABI_NAMESPACE_CLOSE_I_
}} // namespace ztd::text

#endif // ZTD_TEXT_IS_LOSSLESS_HPP
